import urwid
import math
import time

from academictorrents import *
import pdb
UPDATE_INTERVAL = 0.2

class GraphView(urwid.WidgetWrap):
    """
    A class responsible for providing the application's interface and
    graph display.
    """
    palette = [
        ('body',         'black',      'light gray', 'standout'),
        ('header',       'white',      'dark red',   'bold'),
        ('screen edge',  'light blue', 'dark cyan'),
        ('main shadow',  'dark gray',  'black'),
        ('line',         'black',      'light gray', 'standout'),
        ('bg background','light gray', 'black'),
        ('bg 1',         'black',      'dark blue', 'standout'),
        ('bg 1 smooth',  'dark blue',  'black'),
        ('bg 2',         'black',      'dark cyan', 'standout'),
        ('bg 2 smooth',  'dark cyan',  'black'),
        ('button normal','light gray', 'dark blue', 'standout'),
        ('button select','white',      'dark green'),
        ('line',         'black',      'light gray', 'standout'),
        ('pg normal',    'white',      'black', 'standout'),
        ('pg complete',  'white',      'dark magenta'),
        ('pg smooth',     'dark magenta','black')
        ]

    graph_samples_per_bar = 10
    graph_num_bars = 5
    graph_offset_per_second = 2

    def __init__(self, controller):
        self.controller = controller
        self.started = True
        self.start_time = None
        self.offset = 0
        self.last_offset = None        
        urwid.WidgetWrap.__init__(self, self.main_window())
        
    def get_offset_now(self):
        if self.start_time is None:
            return 0
        if not self.started:
            return self.offset
        tdelta = time.time() - self.start_time
        return int(self.offset + (tdelta*self.graph_offset_per_second))    

    def update_graph(self, force_update=False):
        
        #o = self.get_offset_now()
        #if o == self.last_offset and not force_update:
            #return False        
        #self.last_offset = o
       
        #d, max_value, repeat = self.controller.get_data( o, r )
        max_value = 100
        repeat = 100
        l = []
        h = self.controller.handles[self.controller.tindex]
        
        if (not h.is_seed()):
            s = h.status()
            prog = s.progress
            peers = h.get_peer_info()
            for p in peers:
                if p.downloading_piece_index >= 0:
                    l.append([0,float(p.downloading_progress)/p.downloading_total])
                else:
                    l.append([0,0])
        else:
            prog = 1
        
        #if ((o//5)%2==0):
            #l.append([0,15])
            #l.append([0,35])
            #l.append([20,0])
            #l.append([0,45])
            #l.append([85,0])
            #l.append([0,100])
            #l.append([100,0])
        #else:
            #l.append([0,10])
            #l.append([0,35])
            #l.append([30,0])
            #l.append([20,45])
            #l.append([8,0])
            #l.append([0,50])
            #l.append([10,0])            
    
        time.sleep(1)   
        self.graph.set_data(l,max_value)

        # also update progress
        
        #prog = float(time.time()%100) /100
        self.animate_progress.set_completion( prog )
        return True

    def on_animate_button(self, button):
        """Toggle started state and button text."""
        if self.started:
            button.set_label("Start")
            self.offset = self.get_offset_now()
            self.started = False
            self.controller.stop_animation()           
        else:
            self.start_time = time.time()            
            h = self.controller.handles[self.controller.tindex]
            s = h.status()
            button.set_label("Started")
            
            if (h.is_paused()):
                button.set_label("Resuming")
                h.resume()
            
            self.controller.animate_graph()           

    def on_reset_button(self, w):
        self.offset = 0
        self.start_time = time.time()
        self.update_graph(True)

    def on_mode_button(self, button, state):
        """Notify the controller of a new mode setting."""
        if state:
            # The new mode is the label of the button
            self.controller.set_mode( button.get_label() )
        self.last_offset = None

    def on_mode_change(self, m):
        """Handle external mode change by updating radio buttons."""
        for rb in self.mode_buttons:
            if rb.get_label() == m:
                rb.set_state(True, do_callback=False)
                break
        self.last_offset = None

    def on_unicode_checkbox(self, w, state):
        self.graph = self.bar_graph( state )
        self.graph_wrap._w = self.graph
        self.animate_progress = self.progress_bar( state )
        self.animate_progress_wrap._w = self.animate_progress
        self.update_graph( True )


    def main_shadow(self, w):
        """Wrap a shadow and background around widget w."""
        bg = urwid.AttrWrap(urwid.SolidFill(u"\u2592"), 'screen edge')
        shadow = urwid.AttrWrap(urwid.SolidFill(u" "), 'main shadow')

        bg = urwid.Overlay( shadow, bg,
            ('fixed left', 3), ('fixed right', 1),
            ('fixed top', 2), ('fixed bottom', 1))
        w = urwid.Overlay( w, bg,
            ('fixed left', 2), ('fixed right', 3),
            ('fixed top', 1), ('fixed bottom', 2))
        return w

    def bar_graph(self, smooth=False):
        satt = None
        if smooth:
            satt = {(1,0): 'bg 1 smooth', (2,0): 'bg 2 smooth'}
        w = urwid.BarGraph(['bg background','bg 1','bg 2'], satt=satt)
        return w

    def button(self, t, fn):
        w = urwid.Button(t, fn)
        w = urwid.AttrWrap(w, 'button normal', 'button select')
        return w

    def radio_button(self, g, l, fn):
        w = urwid.RadioButton(g, l, False, on_state_change=fn)
        w = urwid.AttrWrap(w, 'button normal', 'button select')
        return w

    def progress_bar(self, smooth=False):
        if smooth:
            return urwid.ProgressBar('pg normal', 'pg complete',
                0, 1, 'pg smooth')
        else:
            return urwid.ProgressBar('pg normal', 'pg complete',
                0, 1)

    def exit_program(self, w):
        raise urwid.ExitMainLoop()

    def graph_controls(self):
        modes = self.controller.get_modes()
        # setup mode radio buttons
        self.mode_buttons = []
        group = []
        for m in modes:
            rb = self.radio_button( group, m, self.on_mode_button )
            self.mode_buttons.append( rb )
        # setup animate button
        self.animate_button = self.button( "", self.on_animate_button)
        self.on_animate_button( self.animate_button )
        self.offset = 0
        self.animate_progress = self.progress_bar()
        animate_controls = urwid.GridFlow( [
            self.animate_button,
            self.button("Details", self.on_reset_button),
            ], 11, 2, 0, 'center')

        if urwid.get_encoding_mode() == "utf8":
            unicode_checkbox = urwid.CheckBox(
                "Enable Unicode Graphics",
                on_state_change=self.on_unicode_checkbox)
        else:
            unicode_checkbox = urwid.Text(
                "UTF-8 encoding not detected")

        self.animate_progress_wrap = urwid.WidgetWrap(
            self.animate_progress)

        l = [    urwid.Text("Torrent Files",align="center"),
            ] + self.mode_buttons + [
            urwid.Divider(),
            urwid.Text("Actions",align="center"),
            animate_controls,
            self.animate_progress_wrap,
            urwid.Divider(),
            urwid.LineBox( unicode_checkbox ),
            urwid.Divider(),
            self.button("Quit", self.exit_program ),
            ]
        w = urwid.ListBox(urwid.SimpleListWalker(l))
        return w

    def main_window(self):
        self.graph = self.bar_graph()
        self.graph_wrap = urwid.WidgetWrap( self.graph )
        vline = urwid.AttrWrap( urwid.SolidFill(u'\u2502'), 'line')
        c = self.graph_controls()
        w = urwid.Columns([('weight',2,self.graph_wrap),
            ('fixed',1,vline), c],
            dividechars=1, focus_column=2)
        w = urwid.Padding(w,('fixed left',1),('fixed right',0))
        w = urwid.AttrWrap(w,'body')
        w = urwid.LineBox(w)
        w = urwid.AttrWrap(w,'line')
        w = self.main_shadow(w)
        return w
    
class GraphController:
    """
    A class responsible for setting up the model and view and running
    the application.
    """
    
    def __init__(self, torrentFiles):
        self.animate_alarm = None
        self.client = Client()
        self.handles = self.client.add_torrent(torrentFiles)
        self.view = GraphView(self)
        # use the first mode as the default
        mode = self.get_modes()[0]
        self.tindex = 0
        # update the view
        self.view.on_mode_change( mode )
        self.view.update_graph(True)

    def get_modes(self):
        """Allow our view access to the list of modes."""
        return self.client.getmodes(self.handles)

    def set_mode(self, m):
        """Allow our view to set the mode."""
        self.tindex = self.client.getmodes(self.handles).index(m)
        self.view.update_graph(True)
        return self.tindex

    def get_data(self, offset, range):
        """Provide data to our view for the graph."""
        return self.client.get_data( offset, range )

    def main(self):
        self.loop = urwid.MainLoop(self.view, self.view.palette)
        self.loop.run()

    def animate_graph(self, loop=None, user_data=None):
        """update the graph and schedule the next update""" 
        self.view.update_graph()
        self.animate_alarm = self.loop.set_alarm_in(
            UPDATE_INTERVAL, self.animate_graph)

    def stop_animation(self):
        """stop animating the graph"""
        if self.animate_alarm:
            self.loop.remove_alarm(self.animate_alarm)
        self.animate_alarm = None


#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Urwid python compatibility definitions
#    Copyright (C) 2011  Ian Ward
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Urwid web site: http://excess.org/urwid/

import sys

try: # python 2.4 and 2.5 compat
    bytes = bytes
except NameError:
    bytes = str

PYTHON3 = sys.version_info > (3, 0)

# for iterating over byte strings:
# ord2 calls ord in python2 only
# chr2 converts an ordinal value to a length-1 byte string
# B returns a byte string in all supported python versions
# bytes3 creates a byte string from a list of ordinal values
if PYTHON3:
    ord2 = lambda x: x
    chr2 = lambda x: bytes([x])
    B = lambda x: x.encode('iso8859-1')
    bytes3 = bytes
else:
    ord2 = ord
    chr2 = chr
    B = lambda x: x
    bytes3 = lambda x: bytes().join([chr(c) for c in x])





#Academic Torrents: Set downloading priorities to files within a torrent

#Copyright (C) [2017] Ronald D. Barrios

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

""" NpyScreen console interface for setting downloading priorities """

import npyscreen

glb_handles = None
glb_so = None


class PriorityForm(npyscreen.FormMultiPageAction):
    """ Class adds support for multi-page forms. By default, scrolling down off
    the last widget on a page moves to the next page, and moving up from the
    first widget moves back a page."""

    def afterEditing(self):
        """ Called when the form is exited.

        Based in the priorityDict obtained in the create() method which contains
        just the position index of the selected items, a new_priorityList is
        created by matching the index positions and placing 1 to download when
        there is match and 0 to not download when there is no match. Finally,
        this new_priorityList gets inserted again in the prioryDict dictionary.
        """

        self.parentApp.setNextForm(None)
        tempDict = self.priorityDict
        for h, priorityList in tempDict.items():
            new_priorityList = []
            for i in range(h.get_torrent_info().num_files()):
                if i in priorityList.value:
                    new_priorityList.append(1)
                else:
                    new_priorityList.append(0)
            self.priorityDict[h] = new_priorityList
        global glb_so
        glb_so = self.priorityDict
        #self.editing = False

    def on_ok(self):
        """ Called when the ok button is pressed."""
        pass

    def on_cancel(self): # pragma: no cover
        """ Called when the cancel button is pressed."""
        exit()

    def create(self):
        """ This method is called by the Form\'s constructor.

        It creates the priorityDict dictionary from the torrent info and from
        the user input through the add_widget_intelligent() method.
        npyscreen.TitleMultiselect offers the user a list of options, allow
        him or her to select more than one of them. The value attribute is
        a list of the indexes user's choices. The list of choices is stored
        in the attribue values.
        """

        self.priorityDict = {}
        for h in glb_handles:
            paths = []
            for f in h.get_torrent_info().files():
                paths.append(f.path)
            self.priorityDict[h] = self.add_widget_intelligent( \
                npyscreen.TitleMultiSelect, max_height=5, value=[1,],
                name=h.name(), values=paths, scroll_exit=True)


class PriorityApp(npyscreen.NPSAppManaged):
    """ NPSAppManaged manages the application."""

    def onStart(self):
        """ Performs initialization of PriorityForm. """

        self.addForm('MAIN', PriorityForm, name='Files to download')

    def set_handler(self, handles):
        """ Sets the list of torrent handles to a global list variable.

        Args:
            handles: a list of libtorrent torrent handler objects.
        """

        global glb_handles
        glb_handles = handles


def get_so():
    """ Returns the global dictionary variable. """

    return glb_so



# Urwid example lazy directory browser / tree view
#    Copyright (C) 2004-2011  Ian Ward
#    Copyright (C) 2010  Kirk McDonald
#    Copyright (C) 2010  Rob Lanphier
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Urwid web site: http://excess.org/urwid/

"""
Urwid example lazy directory browser / tree view
Features:
- custom selectable widgets for files and directories
- custom message widgets to identify access errors and empty directories
- custom list walker for displaying widgets in a tree fashion
- outputs a quoted list of files and directories "selected" on exit
"""
from __future__ import absolute_import
import itertools
import re
import os
import six
import urwid
from menu import SubMenu, Choice, singleton, HorizontalBoxes, getchosen
from academictorrents import Client


class FlagFileWidget(urwid.TreeWidget):

    # apply an attribute to the expand/unexpand icons
    unexpanded_icon = urwid.AttrMap(urwid.TreeWidget.unexpanded_icon, 'dirmark')
    expanded_icon = urwid.AttrMap(urwid.TreeWidget.expanded_icon, 'dirmark')

    def __init__(self, node):
        self.__super.__init__(node)
        # insert an extra AttrWrap for our own use
        self._w = urwid.AttrWrap(self._w, None)
        self.flagged = False
        self.update_w()

    def selectable(self):
        return True

    def keypress(self, size, key):
        """allow subclasses to intercept keystrokes"""
        key = self.__super.keypress(size, key)
        if key:
            key = self.unhandled_keys(size, key)
        return key

    def unhandled_keys(self, size, key):
        """
        Override this method to intercept keystrokes in subclasses.
        Default behavior: Toggle flagged on space, ignore other keys.
        """
        if key == " ":
            self.flagged = not self.flagged
            self.update_w()
        else:
            return key

    def update_w(self):
        """Update the attributes of self.widget based on self.flagged.
        """
        if self.flagged:
            self._w.attr = 'flagged'
            self._w.focus_attr = 'flagged focus'
        else:
            self._w.attr = 'body'
            self._w.focus_attr = 'focus'


class FileTreeWidget(FlagFileWidget):
    """Widget for individual files."""
    def __init__(self, node):
        self.__super.__init__(node)
        path = node.get_value()
        add_widget(path, self)

    def get_display_text(self):
        return self.get_node().get_key()



class EmptyWidget(urwid.TreeWidget):
    """A marker for expanded directories with no contents."""
    def get_display_text(self):
        return ('flag', '(empty directory)')


class ErrorWidget(urwid.TreeWidget):
    """A marker for errors reading directories."""

    def get_display_text(self):
        return ('error', "(error/permission denied)")


class DirectoryWidget(FlagFileWidget):
    """Widget for a directory."""
    def __init__(self, node):
        self.__super.__init__(node)
        path = node.get_value()
        add_widget(path, self)
        self.expanded = starts_expanded(path)
        self.update_expanded_icon()

    def get_display_text(self):
        node = self.get_node()
        if node.get_depth() == 0:
            return "/"
        return node.get_key()


class FileNode(urwid.TreeNode):
    """Metadata storage for individual files"""

    def __init__(self, path, parent=None):
        depth = path.count(dir_sep())
        key = os.path.basename(path)
        urwid.TreeNode.__init__(self, path, key=key, parent=parent, depth=depth)

    def load_parent(self):
        parentname, myname = os.path.split(self.get_value())
        parent = DirectoryNode(parentname)
        parent.set_child_node(self.get_key(), self)
        return parent

    def load_widget(self):
        return FileTreeWidget(self)


class EmptyNode(urwid.TreeNode):
    def load_widget(self):
        return EmptyWidget(self)


class ErrorNode(urwid.TreeNode):
    def load_widget(self):
        return ErrorWidget(self)


class DirectoryNode(urwid.ParentNode):
    """Metadata storage for directories"""

    def __init__(self, path, parent=None):
        if path == dir_sep():
            depth = 0
            key = None
        else:
            depth = path.count(dir_sep())
            key = os.path.basename(path)
        urwid.ParentNode.__init__(self, path, key=key, parent=parent,
                                  depth=depth)

    def load_parent(self):
        parentname, myname = os.path.split(self.get_value())
        parent = DirectoryNode(parentname)
        parent.set_child_node(self.get_key(), self)
        return parent

    def load_child_keys(self):
        dirs = []
        files = []
        try:
            path = self.get_value()
            # separate dirs and files
            for a in os.listdir(path):
                if os.path.isdir(os.path.join(path, a)):
                    dirs.append(a)
                else:
                    files.append(a)
        except OSError, e:
            depth = self.get_depth() + 1
            self._children[None] = ErrorNode(self, parent=self, key=None,
                                             depth=depth)
            return [None]

        # sort dirs and files
        dirs.sort(key=alphabetize)
        files.sort(key=alphabetize)
        # store where the first file starts
        self.dir_count = len(dirs)
        # collect dirs and files together again
        keys = dirs + files
        if len(keys) == 0:
            depth = self.get_depth() + 1
            self._children[None] = EmptyNode(self, parent=self, key=None,
                                             depth=depth)
            keys = [None]
        return keys

    def load_child_node(self, key):
        """Return either a FileNode or DirectoryNode"""
        index = self.get_child_index(key)
        if key is None:
            return EmptyNode(None)

        path = os.path.join(self.get_value(), key)
        if index < self.dir_count:
            return DirectoryNode(path, parent=self)

        path = os.path.join(self.get_value(), key)
        return FileNode(path, parent=self)

    def load_widget(self):
        return DirectoryWidget(self)


class DirectoryBrowser(object):
    palette = [
        ('body', 'black', 'light gray'),
        ('flagged', 'black', 'dark green', ('bold', 'underline')),
        ('focus', 'light gray', 'dark blue', 'standout'),
        ('flagged focus', 'yellow', 'dark cyan',
         ('bold', 'standout', 'underline')),
        ('head', 'yellow', 'black', 'standout'),
        ('foot', 'light gray', 'black'),
        ('key', 'light cyan', 'black', 'underline'),
        ('title', 'white', 'black', 'bold'),
        ('dirmark', 'black', 'dark cyan', 'bold'),
        ('flag', 'dark gray', 'light gray'),
        ('error', 'dark red', 'light gray'),
        ]

    footer_text = [
        ('title', "Directory Browser"), "    ",
        ('key', "UP"), ",", ('key', "DOWN"), ",",
        ('key', "PAGE UP"), ",", ('key', "PAGE DOWN"),
        "  ",
        ('key', "SPACE"), "  ",
        ('key', "+"), ",",
        ('key', "-"), "  ",
        ('key', "LEFT"), "  ",
        ('key', "HOME"), "  ",
        ('key', "END"), "  ",
        ('key', "Q"),
        ]


    def __init__(self):
        cwd = os.getcwd()
        store_initial_cwd(cwd)
        self.header = urwid.Text("")
        self.listbox = urwid.TreeListBox(urwid.TreeWalker(DirectoryNode(cwd)))
        self.listbox.offset_rows = 1
        self.footer = urwid.AttrWrap(urwid.Text(self.footer_text), 'foot')
        self.view = urwid.Frame(
            urwid.AttrWrap(self.listbox, 'body'),
            header=urwid.AttrWrap(self.header, 'head'),
            footer=self.footer)

    def main(self):
        """Run the program."""

        self.loop = urwid.MainLoop(self.view, self.palette,
                                   unhandled_input=self.unhandled_input)
        self.loop.run()

        # on exit, write the flagged filenames to the console
        #names = [escape_filename_sh(x) for x in get_flagged_names()]
        names = get_flagged_names()
        #pdb.set_trace()
        return names

    def unhandled_input(self, k):
        # update display of focus directory
        if k in ('q', 'Q'):
            raise urwid.ExitMainLoop()

class NavigationMenu(object):

    palette = [
        (None, 'light gray', 'black'),
        ('heading', 'black', 'light gray'),
        ('line', 'black', 'light gray'),
        ('options', 'dark gray', 'black'),
        ('focus heading', 'white', 'dark red'),
        ('focus line', 'black', 'dark red'),
        ('focus options', 'black', 'light gray'),
        ('selected', 'white', 'dark blue')]

    menu_top = SubMenu(six.text_type('Main Menu'), [
        SubMenu(six.text_type('Add Torrent'), [
            Choice(six.text_type('from .torrent')),
            Choice(six.text_type('from URL')),
            Choice(six.text_type('from infohash')),
            ]),
        SubMenu(six.text_type('System'), [
            SubMenu(six.text_type('Preferences'), [
                Choice(six.text_type('Appearance')),
                ]),
            Choice(six.text_type('Lock Screen')),
            ]),
    ])

    def __init__(self):
        self.top = singleton(HorizontalBoxes)
        self.top.open_box(self.menu_top.menu)

    def main(self):
        """Run the Navigation Menu"""
        self.loop = urwid.MainLoop(urwid.Filler(self.top,
                                                'middle', 10), self.palette)
        self.loop.run()
        return getchosen()

class Presentation(object):

    palette = [
        ('banner', '', '', '', '#ffa', '#60d'),
        ('streak', '', '', '', 'g50', '#60a'),
        ('inside', '', '', '', 'g38', '#808'),
        ('outside', '', '', '', 'g27', '#a06'),
        ('bg', '', '', '', 'g7', '#d06'), ]

    def __init__(self):
        self.placeholder = urwid.SolidFill()
        self.div = urwid.Divider()
        self.outside = urwid.AttrMap(self.div, 'outside')
        self.inside = urwid.AttrMap(self.div, 'inside')
        self.txt = urwid.Text(('banner',
                               six.text_type('Academic Torrents')),
                              align='center')
        self.streak = urwid.AttrMap(self.txt, 'streak')

    def main(self):
        """Run the cover page"""

        self.loop = urwid.MainLoop(self.placeholder, self.palette,
                                   unhandled_input=self.exit_presentation)
        self.loop.screen.set_terminal_properties(colors=256)
        self.loop.widget = urwid.AttrMap(self.placeholder, 'bg')
        self.loop.widget.original_widget = urwid.Filler(urwid.Pile([]))
        self.pile = self.loop.widget.base_widget

        for item in [self.outside, self.inside, self.streak,
                     self.inside, self.outside]:
            self.pile.contents.append((item, self.pile.options()))
        self.loop.run()

    def exit_presentation(self, k):
        raise urwid.ExitMainLoop()


def main(): # pragma: no cover
    Presentation().main()
    selected_option = NavigationMenu().main()
    if selected_option == six.text_type('from .torrent'):
        torrentFiles = DirectoryBrowser().main()
        myclient = Client()
        handles = myclient.add_torrent(torrentFiles)
        myclient.run_set_priorities(handles)
        myclient.process_handles(handles)

#######
# global cache of widgets
_widget_cache = {}


def add_widget(path, widget):
    """Add the widget for a given path"""

    _widget_cache[path] = widget


def get_flagged_names():
    """Return a list of all filenames marked as flagged."""

    l = []
    for w in _widget_cache.values():
        if w.flagged:
            l.append(w.get_node().get_value())
    return l



######
# store path components of initial current working directory
_initial_cwd = []


def store_initial_cwd(name):
    """Store the initial current working directory path components."""

    global _initial_cwd
    _initial_cwd = name.split(dir_sep())


def starts_expanded(name):
    """Return True if directory is a parent of initial cwd."""

    if name is '/':
        return True

    l = name.split(dir_sep())
    if len(l) > len(_initial_cwd):
        return False

    if l != _initial_cwd[:len(l)]:
        return False

    return True


def escape_filename_sh(name): # pragma: no cover
    """Return a hopefully safe shell-escaped version of a filename."""

    # check whether we have unprintable characters
    for ch in name:
        if ord(ch) < 32:
            # found one so use the ansi-c escaping
            return escape_filename_sh_ansic(name)

    # all printable characters, so return a double-quoted version
    name.replace('\\', '\\\\')
    name.replace('"', '\\"')
    name.replace('`', '\\`')
    name.replace('$', '\\$')
    return '"'+name+'"'
    #return name


def escape_filename_sh_ansic(name): # pragma: no cover
    """Return an ansi-c shell-escaped version of a filename."""

    out = []
    # gather the escaped characters into a list
    for ch in name:
        if ord(ch) < 32:
            out.append("\\x%02x"% ord(ch))
        elif ch == '\\':
            out.append('\\\\')
        else:
            out.append(ch)

    # slap them back together in an ansi-c quote  $'...'
    return "$'" + "".join(out) + "'"

SPLIT_RE = re.compile(r'[a-zA-Z]+|\d+')


def alphabetize(s):
    L = []
    for isdigit, group in itertools.groupby(SPLIT_RE.findall(s),
                                            key=lambda x: x.isdigit()):
        if isdigit:
            for n in group:
                L.append(('', int(n)))
        else:
            L.append((''.join(group).lower(), 0))
    return L


def dir_sep():
    """Return the separator used in this os."""
    return getattr(os.path, 'sep', '/')


if __name__ == "__main__": # pragma: no cover
    main()



#Academic Torrents API

#Copyright (C) [2017] Ronald D. Barrios

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

"""
Academic Torrents Application Programming Interface
Features:
- Initialize a libtorrent session by instantiating an object of the Client Class.
- Add torrents to the libtorrent session.
- Query the torrent handler for downloading information of a single torrent.
- Process the torrent handler for downloading multiple torrents/
- Straightforward get function for simple direct download of a single torrent
  file with no information about subfiles.
- Set priorities for downloading in case of multiple files within a single
  torrent.
"""

from __future__ import absolute_import
import os
import warnings
import time
import sys
import six
from prior import PriorityApp, get_so

try:
    import curses
except ImportError:
    if os.name == 'nt':
        pass

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import libtorrent as lt


class Client(object):
    """ Class that initializes a libtorrent session.

    Attributes:
        ses: An object that represents the libtorrent session
        defaults: A dictionary that keeps default values for the session
    """

    def __init__(self, port=6881, max_download_rate=0, max_upload_rate=0,
                 save_path='.', proxy_host=''):
        """ Contructor that initializes the libtorrent session.

        Sets the ports range whereby downloaded data will flow. It does
        the conversion of the max upload and download rate from KB/s
        to B/s and inserts these in the dictionary of default values.
        After that, those default values are added to the session. Same
        applies to proxy connection and save_path.

        Args:
            port: an int in the range of 0 to 65525 to make a range up to 65535.
            max_download_rate: an int that assigns the maximum rate of
                downloading in KB/s. 0 means infinite.
            max_upload_rate: an int that assigns the maximum rate of
                uploading in KB/s. 0 means infinite.
            save_path: a string that specifies the destination location in the
            file system of the downloaded files.
            proxy_host: a string with information about the http proxy and its
            port to use separated by a colon.
        """

        self.ses = lt.session()
        self.defaults = {'port': 6881}

        if port < 0 or port > 65525:
            self.ses.listen_on(self.defaults['port'], self.defaults['port'] + 10)
        else:
            self.ses.listen_on(port, port+10)
            self.defaults['port'] = port

        max_upload_rate *= 1000
        max_download_rate *= 1000

        if max_upload_rate <= 0:
            max_upload_rate = -1
        if max_download_rate <= 0:
            max_download_rate = -1

        if proxy_host != '':
            ps = lt.proxy_settings()
            ps.type = lt.proxy_type.http
            ps.hostname = proxy_host.split(':')[0]
            ps.port = int(proxy_host.split(':')[1])
            self.ses.set_proxy(ps)

        self.defaults['max_download_rate'] = max_download_rate
        self.defaults['max_upload_rate'] = max_upload_rate
        self.defaults['save_path'] = save_path
        self.defaults['proxy_host'] = proxy_host

        self.ses.set_download_rate_limit(int(max_download_rate))
        self.ses.set_upload_rate_limit(int(max_upload_rate))
        self.ses.set_settings(lt.session_settings())
        self.ses.set_alert_mask(0xfffffff)

    def add_torrent(self, torrent_files):
        """ Adds torrent files to the libtorrent session.

        Receives a list that contains various paths of torrents files and creates
        a dictionary to store the downloading configuration. It resumes the
        downloading process in case of a previous download job that did not
        finish or that actually finished by updating the resume_data in the
        atp configuration dictionary. Creates a torrent handler h for querying
        information that is then appended to a handles list.

        Args:
            torrent_files: a list of strings containing the path to torrent files
                inside the local file system or as an http link in a website or
                as a magnet link.

        Returns:
            a list of torrent handlers objects suitable for querying progress
            information about the torrent while downloading or after it while
            seeding.
        """

        handles = []
        for tf in torrent_files:
            atp = {}
            atp['save_path'] = self.defaults['save_path']
            atp['storage_mode'] = lt.storage_mode_t.storage_mode_sparse
            atp['paused'] = False
            atp['auto_managed'] = True
            atp['duplicate_is_error'] = True

            if (tf.startswith('magnet:') or tf.startswith('http://') or
                    tf.startswith('https://')):
                atp['url'] = tf
            else:
                info = lt.torrent_info(tf)

                try:
                    resume_path = os.path.join(self.defaults['save_path'],
                                               info.name() + '.fastresume')
                    atp['resume_data'] = open(resume_path, 'rb').read()
                except:
                    pass

                atp['ti'] = info

            h = self.ses.add_torrent(atp)
            h.set_max_connections(60)
            h.set_max_uploads(-1)
            handles.append(h)
        return handles

    def query_handle_sp(self, h):
        """ Queries a single torrent handler object.

        Prints in terminal an updated row with download information: progress,
        download_rate, upload_rate, number of peers, number of seeds, number of
        distributed copies and a state. The state can take any value of the
        following: queued, checking, downloading metadata, downloading, finished,
        seeding, allocating and checking fastresume. In case there are alerts to
        pop from libtorrent to be displayed, they are also printed. After data
        is retrieved, application sleeps for 1 s and repeats until downloading
        is finished.

        Args:
            h: a libtorrent torrent handler object.
        """

        while not h.is_finished():
            s = h.status()

            state_str = ['queued', 'checking', 'downloading metadata',
                         'downloading', 'finished', 'seeding', 'allocating',
                         'checking fastresume']
            six.print_('\r%.2f%% complete (down: %.1f kB/s up: %.1f kB/s peers: '
                       '%d seeds: %d distributed copies: %d) %s' %
                       (s.progress * 100, s.download_rate / 1000,
                        s.upload_rate / 1000, s.num_peers, s.num_seeds,
                        s.distributed_copies, state_str[s.state].ljust(20)),
                       end=' ')

            alerts = self.ses.pop_alerts()
            for a in alerts:
                if a.category() & lt.alert.category_t.error_notification:
                    six.print_(a)

            sys.stdout.flush()
            time.sleep(1)

        six.print_('\n', h.name(), 'complete')

    def process_handles(self, handles):
        """ Queries a list with multiple libtorrent torrent handlers.

        By using the curses library, this method prints multiline information
        about each torrent handler (one line per handler object). Data retrieved
        from the query is the same as in the case of the query_handle_sp method.
        This method sleeps for one second as well and repeats the process.

        Args:
            handles: a list of libtorrent torrent handler objects.
        """

        state_str = ['queued', 'checking', 'downloading metadata',
                     'downloading', 'finished', 'seeding',
                     'allocating', 'checking fastresume']
        stdscr = curses.initscr()
        while any(not hand.is_finished() for hand in handles):
            stdscr.erase()
            for h in handles:
                s = h.status()
                stdscr.addstr("{name} {progress}{padding}{percent}% "
                              "complete(down: {down} kB/s up: {up} kB/s peers: "
                              "{peers} seeds: {seeds} copies: "
                              "{distributed_copies}) {state}\n".format(
                                  name=h.name(),
                                  progress="#" * int(s.progress*10),
                                  padding=" " * (10 - int(s.progress*10)),
                                  percent=round(s.progress*100, 2),
                                  down=s.download_rate / 1000,
                                  up=s.upload_rate / 1000,
                                  peers=s.num_peers,
                                  seeds=s.num_seeds,
                                  distributed_copies=round(s.distributed_copies),
                                  state=state_str[s.state]
                                  ))
                alerts = self.ses.pop_alerts()
                for a in alerts:
                    if a.category() & lt.alert.category_t.error_notification:
                        if isinstance(a, str):
                            stdscr.addstr(a)
                        else:
                            pass
                time.sleep(1)
            stdscr.refresh()
        curses.endwin()

    def getmodes(self, handles):
        """ Gets a list with the filenames or top level root directories.

        By receiving a list of torrent handlers, the method queries the name for
        each element of the list. If the target to download is a single file,
        the filename will be retrieved. If the target is a directory containing
        multiple files, then the directory name is retrieved. Each name is
        appended to a list which is returned finalizing the method.

        Ags:
            handles: a list of libtorrent torrent handler objects.

        Returns:
            A list with the target names to download
        """

        names = []
        for h in handles:
            if h.has_metadata():
                names.append(h.name())
            else:
                names.append('-')
        return names

    def getpriorities(self, h):
        """ Returns a list with the priorities of all files within the target.

        There are 8 different priority levels:
        0. piece is not downloaded at all
        1. normal priority. Download order is dependent on availability.
        2. higher than normal priority. Pieces are preferred over pieces with
           the same availability, but not over pieces with lower availability.
        3. pieces are as likely to be picked as partial pieces.
        4. pieces are preferred over partial pieces, but not over pieces with
           lower availability.
        5. currently the same as 4.
        6. piece is as likely to be picked as any piece with availability 1.
        7. maximum priority, availability is disregarded, the piece is preferred
           over any other piece with lower priority.

        Args:
            h: a libtorrent torrent handler object.

        Returns:
            A list with the priorities of all files within a single torrent.
        """

        if h.has_metadata():
            return h.file_priorities()
        return 0

    def setpriorities(self, h, priorityList):
        """ Sets the downloading priority of all files.

        Args:
            h: a libtorrent torrent handler object.
            priorityList: a list with int elements in the range 0-7. See full
            specification in the getpriorities method.
        """

        if h.has_metadata():
            h.prioritize_files(priorityList)

    def prioritize_file(self, h, fileIndex, priorityIndex):
        """ Sets the priority index for just one of the files.

        Args:
            h: a libtorrent torrent handler object.
            fileIndex: the index position of the file which priority is desired
            to change.
            priorityIndex: an int in the range 0-7.
        """

        if h.has_metadata():
            h.file_priority(fileIndex, priorityIndex)

    def run_set_priorities(self, handles):
        """ Runs npyscreen application to set the priorities of various torrent
        handlers.

        Args:
            handles: a list of libtorrent torrent handler objects.
        """

        if any(hand.get_torrent_info().num_files() > 1 for hand in handles):
            pApp = PriorityApp()
            pApp.set_handler(handles)
            pApp.run()

            for h, pl in get_so().items():
                if h.has_metadata():
                    h.prioritize_files(pl)


def get(tf):
    """ Downloads a single torrent file by a simple get() client method.

    No need for the API user to write bolierplate code for Client() object
    instantiation, torrent adding and name retrieving.

    Args:
        tf: a string that represents the path of the .torrent file.
    """

    client = Client()
    handles = client.add_torrent([tf])
    filenames = client.getmodes(handles)
    client.query_handle_sp(handles[0])
    return filenames[0]


def version():
    """ Returns App Version Number."""

    return 'Academic Torrents Version 1.8'


def release_info():
    """ Prints environment variable information and application versioning."""

    six.print_('LD_LIBRARY PATH set to: ', os.environ['LD_LIBRARY_PATH'])
    six.print_('PKG_CONFIG_PATH set to: ', os.environ['PKG_CONFIG_PATH'])
    six.print_('Libtorrent Version Installed is: ' + lt.version)
    six.print_(version())


if __name__ == "__main__": # pragma: no cover
    release_info()



# Urwid terminal emulation widget unit tests
#    Copyright (C) 2010  aszlig
#    Copyright (C) 2011  Ian Ward
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Urwid web site: http://excess.org/urwid/

import os
import sys
import unittest

from itertools import dropwhile

from urwid import vterm
from urwid import signals
from urwid.compat import B


class DummyCommand(object):
    QUITSTRING = B('|||quit|||')

    def __init__(self):
        self.reader, self.writer = os.pipe()

    def __call__(self):
        # reset
        stdout = getattr(sys.stdout, 'buffer', sys.stdout)
        stdout.write(B('\x1bc'))

        while True:
            data = os.read(self.reader, 1024)
            if self.QUITSTRING == data:
                break
            stdout.write(data)
            stdout.flush()

    def write(self, data):
        os.write(self.writer, data)

    def quit(self):
        self.write(self.QUITSTRING)


class TermTest(unittest.TestCase):
    def setUp(self):
        self.command = DummyCommand()

        self.term = vterm.Terminal(self.command)
        self.resize(80, 24)

    def tearDown(self):
        self.command.quit()

    def connect_signal(self, signal):
        self._sig_response = None

        def _set_signal_response(widget, *args, **kwargs):
            self._sig_response = (args, kwargs)
        self._set_signal_response = _set_signal_response

        signals.connect_signal(self.term, signal, self._set_signal_response)

    def expect_signal(self, *args, **kwargs):
        self.assertEqual(self._sig_response, (args, kwargs))

    def disconnect_signal(self, signal):
        signals.disconnect_signal(self.term, signal, self._set_signal_response)

    def caught_beep(self, obj):
        self.beeped = True

    def resize(self, width, height, soft=False):
        self.termsize = (width, height)
        if not soft:
            self.term.render(self.termsize, focus=False)

    def write(self, data):
        data = B(data)
        self.command.write(data.replace(B('\e'), B('\x1b')))

    def flush(self):
        self.write(chr(0x7f))

    def read(self, raw=False):
        self.term.wait_and_feed()
        rendered = self.term.render(self.termsize, focus=False)
        if raw:
            is_empty = lambda c: c == (None, None, B(' '))
            content = list(rendered.content())
            lines = [list(dropwhile(is_empty, reversed(line)))
                     for line in content]
            return [list(reversed(line)) for line in lines if len(line)]
        else:
            content = rendered.text
            lines = [line.rstrip() for line in content]
            return B('\n').join(lines).rstrip()

    def expect(self, what, desc=None, raw=False):
        if not isinstance(what, list):
            what = B(what)
        got = self.read(raw=raw)
        if desc is None:
            desc = ''
        else:
            desc += '\n'
        desc += 'Expected:\n%r\nGot:\n%r' % (what, got)
        self.assertEqual(got, what, desc)

    def test_simplestring(self):
        self.write('hello world')
        self.expect('hello world')

    def test_linefeed(self):
        self.write('hello\x0aworld')
        self.expect('hello\nworld')

    def test_linefeed2(self):
        self.write('aa\b\b\eDbb')
        self.expect('aa\nbb')

    def test_carriage_return(self):
        self.write('hello\x0dworld')
        self.expect('world')

    def test_insertlines(self):
        self.write('\e[0;0flast\e[0;0f\e[10L\e[0;0ffirst\nsecond\n\e[11D')
        self.expect('first\nsecond\n\n\n\n\n\n\n\n\nlast')

    def test_deletelines(self):
        self.write('1\n2\n3\n4\e[2;1f\e[2M')
        self.expect('1\n4')

    def test_movement(self):
        self.write('\e[10;20H11\e[10;0f\e[20C\e[K')
        self.expect('\n' * 9 + ' ' * 19 + '1')
        self.write('\e[A\e[B\e[C\e[D\b\e[K')
        self.expect('')
        self.write('\e[50A2')
        self.expect(' ' * 19 + '2')
        self.write('\b\e[K\e[50B3')
        self.expect('\n' * 23 + ' ' * 19 + '3')
        self.write('\b\e[K' + '\eM' * 30 + '\e[100C4')
        self.expect(' ' * 79 + '4')
        self.write('\e[100D\e[K5')
        self.expect('5')

    def edgewall(self):
        edgewall = '1-\e[1;%(x)df-2\e[%(y)d;1f3-\e[%(y)d;%(x)df-4\x0d'
        self.write(edgewall % {'x': self.termsize[0] - 1,
                               'y': self.termsize[1] - 1})

    def test_horizontal_resize(self):
        self.resize(80, 24)
        self.edgewall()
        self.expect('1-' + ' ' * 76 + '-2' + '\n' * 22
                         + '3-' + ' ' * 76 + '-4')
        self.resize(78, 24, soft=True)
        self.flush()
        self.expect('1-' + '\n' * 22 + '3-')
        self.resize(80, 24, soft=True)
        self.flush()
        self.expect('1-' + '\n' * 22 + '3-')

    def test_vertical_resize(self):
        self.resize(80, 24)
        self.edgewall()
        self.expect('1-' + ' ' * 76 + '-2' + '\n' * 22
                         + '3-' + ' ' * 76 + '-4')
        for y in xrange(23, 1, -1):
            self.resize(80, y, soft=True)
            self.write('\e[%df\e[J3-\e[%d;%df-4' % (y, y, 79))
            desc = "try to rescale to 80x%d." % y
            self.expect('\n' * (y - 2) + '3-' + ' ' * 76 + '-4', desc)
        self.resize(80, 24, soft=True)
        self.flush()
        self.expect('1-' + ' ' * 76 + '-2' + '\n' * 22
                         + '3-' + ' ' * 76 + '-4')

    def write_movements(self, arg):
        fmt = 'XXX\n\e[faaa\e[Bccc\e[Addd\e[Bfff\e[Cbbb\e[A\e[Deee'
        self.write(fmt.replace('\e[', '\e['+arg))

    def test_defargs(self):
        self.write_movements('')
        self.expect('aaa   ddd      eee\n   ccc   fff bbb')

    def test_nullargs(self):
        self.write_movements('0')
        self.expect('aaa   ddd      eee\n   ccc   fff bbb')

    def test_erase_line(self):
        self.write('1234567890\e[5D\e[K\n1234567890\e[5D\e[1K\naaaaaaaaaaaaaaa\e[2Ka')
        self.expect('12345\n      7890\n               a')

    def test_erase_display(self):
        self.write('1234567890\e[5D\e[Ja')
        self.expect('12345a')
        self.write('98765\e[8D\e[1Jx')
        self.expect('   x5a98765')

    def test_scrolling_region_simple(self):
        self.write('\e[10;20r\e[10f1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n12\e[faa')
        self.expect('aa' + '\n' * 9 + '2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n12')

    def test_scrolling_region_reverse(self):
        self.write('\e[2J\e[1;2r\e[5Baaa\r\eM\eM\eMbbb\nXXX')
        self.expect('\n\nbbb\nXXX\n\naaa')

    def test_scrolling_region_move(self):
        self.write('\e[10;20r\e[2J\e[10Bfoo\rbar\rblah\rmooh\r\e[10Aone\r\eM\eMtwo\r\eM\eMthree\r\eM\eMa')
        self.expect('ahree\n\n\n\n\n\n\n\n\n\nmooh')

    def test_scrolling_twice(self):
        self.write('\e[?6h\e[10;20r\e[2;5rtest')
        self.expect('\ntest')

    def test_cursor_scrolling_region(self):
        self.write('\e[?6h\e[10;20r\e[10f1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n12\e[faa')
        self.expect('\n' * 9 + 'aa\n3\n4\n5\n6\n7\n8\n9\n10\n11\n12')

    def test_relative_region_jump(self):
        self.write('\e[21H---\e[10;20r\e[?6h\e[18Htest')
        self.expect('\n' * 19 + 'test\n---')

    def test_set_multiple_modes(self):
        self.write('\e[?6;5htest')
        self.expect('test')
        self.assertTrue(self.term.term_modes.constrain_scrolling)
        self.assertTrue(self.term.term_modes.reverse_video)
        self.write('\e[?6;5l')
        self.expect('test')
        self.assertFalse(self.term.term_modes.constrain_scrolling)
        self.assertFalse(self.term.term_modes.reverse_video)

    def test_wrap_simple(self):
        self.write('\e[?7h\e[1;%dHtt' % self.term.width)
        self.expect(' ' * (self.term.width - 1) + 't\nt')

    def test_wrap_backspace_tab(self):
        self.write('\e[?7h\e[1;%dHt\b\b\t\ta' % self.term.width)
        self.expect(' ' * (self.term.width - 1) + 'a')

    def test_cursor_visibility(self):
        self.write('\e[?25linvisible')
        self.expect('invisible')
        self.assertEqual(self.term.term.cursor, None)
        self.write('\rvisible\e[?25h\e[K')
        self.expect('visible')
        self.assertNotEqual(self.term.term.cursor, None)

    def test_get_utf8_len(self):
        length = self.term.term.get_utf8_len(int("11110000", 2))
        self.assertEqual(length, 3)
        length = self.term.term.get_utf8_len(int("11000000", 2))
        self.assertEqual(length, 1)
        length = self.term.term.get_utf8_len(int("11111101", 2))
        self.assertEqual(length, 5)

    def test_encoding_unicode(self):
        vterm.util._target_encoding = 'utf-8'
        self.write('\e%G\xe2\x80\x94')
        self.expect('\xe2\x80\x94')

    def test_encoding_unicode_ascii(self):
        vterm.util._target_encoding = 'ascii'
        self.write('\e%G\xe2\x80\x94')
        self.expect('?')

    def test_encoding_wrong_unicode(self):
        vterm.util._target_encoding = 'utf-8'
        self.write('\e%G\xc0\x99')
        self.expect('')

    def test_encoding_vt100_graphics(self):
        vterm.util._target_encoding = 'ascii'
        self.write('\e)0\e(0\x0fg\x0eg\e)Bn\e)0g\e)B\e(B\x0fn')
        self.expect([[
            (None, '0', B('g')), (None, '0', B('g')),
            (None, None, B('n')), (None, '0', B('g')),
            (None, None, B('n'))
        ]], raw=True)

    def test_ibmpc_mapping(self):
        vterm.util._target_encoding = 'ascii'

        self.write('\e[11m\x18\e[10m\x18')
        self.expect([[(None, 'U', B('\x18'))]], raw=True)

        self.write('\ec\e)U\x0e\x18\x0f\e[3h\x18\e[3l\x18')
        self.expect([[(None, None, B('\x18'))]], raw=True)

        self.write('\ec\e[11m\xdb\x18\e[10m\xdb')
        self.expect([[
            (None, 'U', B('\xdb')), (None, 'U', B('\x18')),
            (None, None, B('\xdb'))
        ]], raw=True)

    def test_set_title(self):
        self._the_title = None

        def _change_title(widget, title):
            self._the_title = title

        self.connect_signal('title')
        self.write('\e]666parsed right?\e\\te\e]0;test title\007st1')
        self.expect('test1')
        self.expect_signal(B('test title'))
        self.write('\e]3;stupid title\e\\\e[0G\e[2Ktest2')
        self.expect('test2')
        self.expect_signal(B('stupid title'))
        self.disconnect_signal('title')

    def test_set_leds(self):
        self.connect_signal('leds')
        self.write('\e[0qtest1')
        self.expect('test1')
        self.expect_signal('clear')
        self.write('\e[3q\e[H\e[Ktest2')
        self.expect('test2')
        self.expect_signal('caps_lock')
        self.disconnect_signal('leds')



#Academic Torrents setup.py for installing the software application

#Copyright (C) [2017] Ronald D. Barrios

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

""" Custom setup.py install through the BuildingLibtorrent Class
Features:
- Installs Boost from source code.
- Installs Openssl from source code.
- Installs Libtorrent from source code.
- Generates an .at_profile for sourcing environment variables required by
  main academictorrents application.
- Installs via pip other package dependencies.
"""

from distutils import log
import urllib
import os
import tarfile
import sys
import logging
import subprocess as sub
import shutil as sh
from setuptools import setup
from setuptools.command.install import install
import six


def message(code, phase):
    """ Prints a custom message that depends on the code returned by the
    subprocess command executor.

    Args:
        code: int value returned by subprocess.Popen call.
        phase: a string that describes the context in which subprocess.Popen
               was called.
    """

    if code == 0:
        six.print_(phase + " successful")
    else:
        six.print_(phase+ " with errors. See install.log")


def log_subprocess_output(pipe):
    """ Writes a log file to register what subprocess performed.

    Args:
        pipe: stdout line generated by the subprocess.Popen call
    """

    for line in iter(pipe.readline, b''): # b'\n'-separated lines
        logging.info('%r', line)


class BuildingLibtorrent(install):
    """ Custom install Class that builds libtorrent from source code plus all
    its dependencies to be able to work.

    By providing the --user arg to this setup.py module, the custom installation
    is run. If this argument is not provided, it justs installs the python
    modules.

    Usage:
        Option1: $ pip install --install-option='--user' academictorrents
        Option2: $ python setup.py install -- user
    """

    def run(self):
        origin = os.getcwd()
        if '--user' in sys.argv:
            at_home = os.path.join(os.environ['HOME'], ".local")
            os.chdir(at_home)
            logging.basicConfig(filename='install.log',
                                filemode='w', level=logging.DEBUG)
            pkg = os.path.join(os.environ['HOME'], "at-pkg")
            os.mkdir(pkg)
            os.chdir(pkg)

            six.print_("Preparing Boost")
            boost_url = ('https://sourceforge.net/projects/boost/files/'
                         'boost/1.63.0/boost_1_63_0.tar.gz')
            boost_tb = 'boost_1_63_0.tar.gz'
            boost_dir = 'boost_1_63_0'
            urllib.urlretrieve(boost_url, boost_tb)

            six.print_("Preparing Openssl")
            openssl_url = 'https://www.openssl.org/source/openssl-1.1.0e.tar.gz'
            openssl_tb = 'openssl-1.1.0e.tar.gz'
            openssl_dir = 'openssl-1.1.0e'
            urllib.urlretrieve(openssl_url, openssl_tb)

            six.print_("Preparing Libtorrent")
            libtorrent_url = ('https://github.com/arvidn/libtorrent/'
                              'releases/download/libtorrent-1_1_3/'
                              'libtorrent-rasterbar-1.1.3.tar.gz')
            libtorrent_tb = 'libtorrent-rasterbar-1.1.3.tar.gz'
            libtorrent_dir = 'libtorrent-rasterbar-1.1.3'
            urllib.urlretrieve(libtorrent_url, libtorrent_tb)

            six.print_("Uncompressing libraries")
            for lib in [boost_tb, openssl_tb, libtorrent_tb]:
                six.print_("Uncompressing " + lib)
                tar = tarfile.open(lib)
                tar.extractall()
            tar.close()
            with_python = os.popen('which python2.7').read()
            six.print_("Python executable to use is: " + with_python)

            six.print_("Configuring Openssl")
            os.chdir(os.path.join(pkg, openssl_dir))

            cmd = ['./config', '--prefix=%s'%at_home,
                   '--openssldir=%s'%os.path.join(at_home, "ssl")]

            myproc = sub.Popen(cmd, stdout=sub.PIPE)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Openssl"))

            six.print_("Make Openssl")
            os.chdir(os.path.join(pkg, openssl_dir))
            myproc = sub.Popen(['make'], stdout=sub.PIPE, stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Make Openssl"))

            six.print_("Make test Openssl")
            os.chdir(os.path.join(pkg, openssl_dir))
            myproc = sub.Popen(['make', 'test'], stdout=sub.PIPE,
                               stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Make test Openssl"))

            six.print_("Make install Openssl")
            os.chdir(os.path.join(pkg, openssl_dir))
            myproc = sub.Popen(['make', 'install'], stdout=sub.PIPE,
                               stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode,
                    six.text_type("Make install Openssl"))

            six.print_("Bootstrapping Boost")
            os.chdir(os.path.join(pkg, boost_dir))

            cmd = ['./bootstrap.sh', '--with-python=%s'%with_python,
                   '--prefix=%s'%at_home]

            myproc = sub.Popen(cmd, stdout=sub.PIPE)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Bootstrap"))

            six.print_("Building Boost C++ libraries")
            cmd = ['./b2', '-s', 'NO_BZIP2=1', '--with-system',
                   '--with-date_time', '--with-python', '--with-chrono',
                   '--with-random', 'install', '--prefix=%s'%at_home]

            myproc = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Building"))

            six.print_("Configuring Libtorrent")
            os.chdir(os.path.join(pkg, libtorrent_dir))
            cmd = ['./configure', '--enable-python-binding',
                   '--disable-dependency-tracking', '--disable-silent-rules',
                   'PYTHON=%s'%with_python, '--prefix=%s'%at_home,
                   '--with-boost=%s'%at_home, '--with-libiconv',
                   '--with-boost-python=boost_python',
                   '--with-openssl=%s'%at_home,
                   'BOOST_ROOT=%s'%os.path.join(pkg, boost_dir)]

            myproc = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Configuring Libtorrent"))

            six.print_("Making Libtorrent")
            os.chdir(os.path.join(pkg, libtorrent_dir))
            myproc = sub.Popen(['make'], stdout=sub.PIPE, stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode, six.text_type("Making Libtorrent"))

            six.print_("Make install Libtorrent")
            os.chdir(os.path.join(pkg, libtorrent_dir))
            myproc = sub.Popen(['make', 'install'], stdout=sub.PIPE,
                               stderr=sub.STDOUT)
            with myproc.stdout:
                log_subprocess_output(myproc.stdout)
            myproc.wait()
            message(myproc.returncode,
                    six.text_type("Making install Libtorrent"))

            sh.rmtree(pkg)
            six.print_("Building of all C++ libraries finished")

            six.print_("Creating .at_profile")
            with open(os.path.join(os.environ['HOME'],
                                   ".at_profile"), "w") as outfile:
                outfile.write("export LD_LIBRARY_PATH=" +
                              os.path.join(at_home, "lib") + "\n")
                outfile.write("export PKG_CONFIG_PATH=" +
                              os.path.join(at_home, "lib", "pkgconfig"))

            six.print_("Finished custom install")
            os.chdir(origin)

        install.run(self)
        import pip
        required_packages = self.distribution.install_requires
        for package in required_packages:
            self.announce('Installing %s...' % package, log.INFO)
            pip.main(['install', '--user', package])
        self.announce('Dependencies installed.')

setup(
    name='AcademicTorrents',
    version='1.9',
    author='Ronald Barrios',
    author_email='ronald.degmar@gmail.com',
    packages=['urwid', 'tests'],
    include_package_data=True,
    keywords='torrent bittorrent p2p fileshare filesharing',
    long_description="""AcademicTorrents is a bittorrent client that has
        been developed in order to be used in servers where you cannot sudo.
        AcademicTorrents uses libtorrent in it's backend to handle the
        bittorrent protocol. Console User Interface has been developed by
        using urwid, curses, and npyscreen python libraries""",
    license='GPLv3',
    py_modules=['menu', 'console', 'academictorrents', 'prior'],
    entry_points={
        'console_scripts': ['at-console = console:main'],
    },
    package_data={'tests': ['*.torrent', '.coveragerc']},
    install_requires=[
        'npyscreen', 'coverage',
    ],
    cmdclass={
        'install': BuildingLibtorrent
    }
)



#Academic Torrents Unit Tests

#Copyright (C) [2017] Ronald D. Barrios

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

""" Academic Torrents Unit Tests
Features:
- TestGet Class tests simple at_get() function.
- TestClient Class tests whole Client class plus npyscreen console display.
- TestConsole Class tests urwid console interface.
"""

import unittest
import os
import sys
import six
import academictorrents as at
import console


class TestGet(unittest.TestCase):
    """ This class tests the get function call.

    This is the preferred function to use for downloading a single torrent
    without too much boilerplate code from Client Class.
    """

    def test_download_at_get(self):
        """Try downloading a torrent from Academic Torrents
        and check if it successfully finishes downloading
        until a complete message shows as final output
        """

        filename = at.get('2269d7d1c77375aea732eea0905e370d4741575f.torrent')
        six.print_('\n', 'Check out the downloaded file in ' + os.getcwd(), '\n')
        self.assertEqual(filename, six.text_type('Race-Results.zip'))
        self.assertTrue(os.path.isfile('Race-Results.zip'))


class TestClient(unittest.TestCase):
    """ This class tests various downloading options that are only available in
    the Client Class.
    """

    def test_download_magnet(self):
        """Try downloading a magnet link from Academic Torrents
        and check if it successfully finishes downloading
        until a complete message shows as final output
        """

        # testing with wrong port 80000 which defaults to a valid ports range
        client = at.Client(port=80000)
        magnet_link = ('magnet:?xt=urn:btih:323a0048d87ca79b68f12a6350a57776b6a3'
                       'b7fb&tr=http%3A%2F%2Facademictorrents.com%2Fannounce.php'
                       '&tr=udp%3A%2F%2Ftracker.publicbt.com%3A80%2Fannounce&tr='
                       'udp%3A%2F%2Ftracker.openbittorrent.com%3A80%2Fannounce')
        handles = client.add_torrent([magnet_link])
        filename = client.getmodes(handles)
        # filename is not available in the case of magnet
        # link before downloading is complete
        self.assertEqual(filename[0], six.text_type('-'))
        # Download must be completed first; otherwise, getpriorities
        # will return 0
        self.assertEqual(client.getpriorities(handles[0]), 0)
        client.query_handle_sp(handles[0])
        filename = client.getmodes(handles)
        six.print_('\n', 'Check out the downloaded file in ' + os.getcwd(), '\n')
        self.assertEqual(filename[0], six.text_type('mnist.pkl.gz'))
        self.assertTrue(os.path.isfile('mnist.pkl.gz'))

    def test_download_http(self):
        """Try downloading a torrent from a URL and check if it
        successfully finishes downloading until a complete message
        shows as final output
        """

        # testing with wrong port -50 which defaults to a valid ports range
        client = at.Client(port=-50)
        http_url = ('http://academictorrents.com/download/4cdade8fa3c6d451441fda'
                    '7ae19bb53537b6ee21.torrent')
        handles = client.add_torrent([http_url])
        filename = client.getmodes(handles)
        # filename is not available in the case of http url before downloading
        # is complete.
        self.assertEqual(filename[0], six.text_type('-'))
        # Download must be completed first; otherwise, getpriorities will
        # return 0
        self.assertEqual(client.getpriorities(handles[0]), 0)
        client.query_handle_sp(handles[0])
        filename = client.getmodes(handles)
        six.print_('\n', 'Check out the downloaded file in ' +
                   os.getcwd(), '\n')
        self.assertEqual(filename[0],
                         six.text_type('devinberg.com_471592PB.pdf'))
        self.assertTrue(os.path.isfile('devinberg.com_471592PB.pdf'))

    def test_set_priorities(self):
        """ Try downloading from a torrent file and check if the
        setpriorities and getpriorities() methods set and return
        correct lists of priorities. Also check if the train-labels
        file was correctly downloaded (1 means download and 0 means
        do not download)."""

        client = at.Client()
        torrent_file = '42714f859770f1a9d8b27985f9f16ea17e8ba2f6.torrent'
        handles = client.add_torrent([torrent_file])
        client.setpriorities(handles[0], [0, 1, 0])
        expected_after_set = [0, 1, 0]
        self.assertListEqual(client.getpriorities(handles[0]),
                             expected_after_set)
        client.query_handle_sp(handles[0])
        six.print_('\n', 'Check out the downloaded file in ' + os.getcwd(), '\n')
        self.assertTrue(os.path.isfile(os.path.join(os.getcwd(),
                                                    'files',
                                                    'train-labels.tif')))
        self.assertEquals(os.path.getsize(os.path.join(os.getcwd(), 'files',
                                                       'train-labels.tif')),
                          7869573, msg='download size is incomplete')

    def test_prioritize_file(self):
        """ Try downloading from a torrent file and check if the
        prioritize_file and getpriorities() methods set and return
        correct lists of priorities. Also check if the train-labels
        file was correctly downloaded (1 means download and 0 means
        do not download)."""

        client = at.Client()
        torrent_file = '1d16994c70b7fff8bfe917f83c397b1193daee7f.torrent'
        handles = client.add_torrent([torrent_file])
        client.prioritize_file(handles[0], 0, 0)
        client.prioritize_file(handles[0], 1, 0)
        expected_after_set = [0, 0, 1]
        self.assertListEqual(client.getpriorities(handles[0]),
                             expected_after_set)
        client.query_handle_sp(handles[0])
        six.print_('\n', 'Check out the downloaded file in ' +
                   os.getcwd(), '\n')
        self.assertTrue(os.path.isfile(os.path.join(os.getcwd(),
                                                    'coil-20',
                                                    'coil-20.ps.zip')))
        self.assertEquals(os.path.getsize(os.path.join(os.getcwd(), 'coil-20',
                                                       'coil-20.ps.zip')),
                          442932, msg='download size is incomplete')

    @unittest.skipUnless(sys.platform.startswith('linux'), 'requires Linux')
    def test_process_handles(self):
        """ This is a npyscreen and curses libraries test, so it is
        only available for POSIX systems. The goal is to check if the
        user can visualize actual displayss for establishing downloading
        priorities by using npyscreen library and for downloading
        multiple files by using curses library respectively"""

        client = at.Client()
        torrent_list = ['3efc53f35d49669b89039f2b4ec9de11ec1d73fd.torrent',
                        'ce990b28668abf16480b8b906640a6cd7e3b8b21.torrent']
        handles = client.add_torrent(torrent_list)
        client.run_set_priorities(handles)
        client.process_handles(handles)
        six.print_('\n', 'Check out the downloaded files in ' + os.getcwd(), '\n')

    def test_path_info_and_versioning(self):
        """ This is only a look and check test. The goal here is to make
        sure the environment variables for academictorrents such as
        LD_LIBRARY_PATH and PKG_CONFIG_PATH were correctly set"""

        at.release_info()


class TestConsole(unittest.TestCase):
    """ This class tests urwid menus and browser window."""

    @unittest.skipUnless(sys.platform.startswith('linux'), 'requires Linux')
    def test_at_navigation(self):
        """ Testing of Presentation Cover and Navigation Menu of
        Academic Torrents"""

        console.Presentation().main()
        selected_option = console.NavigationMenu().main()
        pool_of_options = ('from .torrent', 'from URL', 'from infohash',
                           'Appearance', 'Lock Screen')
        self.assertIn(selected_option, pool_of_options,
                      msg='selected option is not in tuple')

    @unittest.skipUnless(sys.platform.startswith('linux'), 'requires Linux')
    def test_at_browser(self):
        """ Testing of Directory Browser of Academic Torrents"""

        torrent_files = console.DirectoryBrowser().main()
        six.print_('These are the selected files:')
        six.print_(torrent_files)


if __name__ == '__main__': # pragma: no cover
    unittest.main()



#Academic Torrents Menu

#Copyright (C) [2017] Ronald D. Barrios

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

""" Horizontal Menu for AcademicTorrents
New menus appear on the right and push old menus off the left side of the screen.
The look of buttons and other menu elements are heavily customized and new
widget classes are used instead of factory functions.
"""

import urwid
import six


class MenuButton(urwid.Button):
    """ MenuButton is a customized Button widget.

    Button uses WidgetWrap to create its appearance and this class replaces
    the display widget created by Button by the wrapped widget in self._w.

    Attributes:
        _w: wraped widget
    """

    def __init__(self, caption, callback):
        super(MenuButton, self).__init__("")
        urwid.connect_signal(self, 'click', callback)
        self._w = urwid.AttrMap(urwid.SelectableIcon(
            [u'  \N{BULLET} ', caption], 2), None, 'selected')


class SubMenu(urwid.WidgetWrap):
    """ SubMenu is implemented with a MenuButton but uses WidgetWrap to hide the
    implementation instead of inheriting from MenuButton.

    The constructor builds a widget for the menu that this button will open and
    stores it in self.menu.
    """

    def __init__(self, caption, choices):
        super(SubMenu, self).__init__(MenuButton(
            [caption, u"\N{HORIZONTAL ELLIPSIS}"], self.open_menu))
        line = urwid.Divider(u'\N{LOWER ONE QUARTER BLOCK}')
        listbox = urwid.ListBox(urwid.SimpleFocusListWalker([
            urwid.AttrMap(urwid.Text([u"\n  ", caption]), 'heading'),
            urwid.AttrMap(line, 'line'),
            urwid.Divider()] + choices + [urwid.Divider()]))
        self.menu = urwid.AttrMap(listbox, 'options')

    def open_menu(self, button):
        """ Open menu options by retrieving the unique instance of
        HorizontalBoxes class"""

        top = singleton(HorizontalBoxes)
        top.open_box(self.menu)


class Choice(urwid.WidgetWrap):
    """ Choice is like SubMenu but displays the item chosen instead of another menu."""

    def __init__(self, caption):
        super(Choice, self).__init__(
            MenuButton(caption, self.item_chosen))
        self.caption = caption

    def item_chosen(self, button):
        """ Saves the selected option and prepares a response upon completion."""

        savechosen(self.caption)
        response = urwid.Text([u'  You chose ', self.caption, u'\n'])
        done = MenuButton(u'Ok', exit_program)
        response_box = urwid.Filler(urwid.Pile([response, done]))
        top = singleton(HorizontalBoxes)
        top.open_box(urwid.AttrMap(response_box, 'options'))


def exit_program(key):
    """ Exits urwid Main Loop"""

    raise urwid.ExitMainLoop()


focus_map = {
    'heading': 'focus heading',
    'options': 'focus options',
    'line': 'focus line'}


class HorizontalBoxes(urwid.Columns):
    """ HorizontalBoxes arranges the menus displayed. There is no special
    handling required for going to previous menus here because Columns already
    handles switching focus when LEFT or RIGHT is pressed. AttrMap with the
    focus_map dict is used to change the appearance of a number of the display
    attributes when a menu is in focus.
    """

    _instance = None

    def __init__(self):
        super(HorizontalBoxes, self).__init__([], dividechars=1)

    def open_box(self, box):
        """ Updates contents"""

        if self.contents:
            del self.contents[self.focus_position + 1:]
        self.contents.append((urwid.AttrMap(box, 'options', focus_map),
                              self.options('given', 24)))
        self.focus_position = len(self.contents) - 1


def singleton(klass):
    """ a method for retrieving a previously created instance. If there is no
    instance, then a new one is created.
    """

    if not klass._instance:
        klass._instance = klass()
    return klass._instance

glb_chosen = six.text_type('')


def savechosen(chosen):
    """ Modifies the _chosen global variable."""

    global glb_chosen
    glb_chosen = chosen


def getchosen():
    """ Returns the modified _chosen global variable."""

    return glb_chosen
